from __future__ import print_function, absolute_import, unicode_literals
from hashlib import sha256
import six
from zope.interface import implementer
from attr import attrs, attrib
from attr.validators import provides, instance_of
from spake2 import SPAKE2_Symmetric
from hkdf import Hkdf
from nacl.secret import SecretBox
from nacl.exceptions import CryptoError
from nacl import utils
from automat import MethodicalMachine
from .util import (to_bytes, bytes_to_hexstr, hexstr_to_bytes,
                   bytes_to_dict, dict_to_bytes)
from . import _interfaces
CryptoError
__all__ = ["derive_key", "derive_phase_key", "CryptoError",
           "Key"]

def HKDF(skm, outlen, salt=None, CTXinfo=b""):
    return Hkdf(salt, skm).expand(CTXinfo, outlen)

def derive_key(key, purpose, length=SecretBox.KEY_SIZE):
    if not isinstance(key, type(b"")): raise TypeError(type(key))
    if not isinstance(purpose, type(b"")): raise TypeError(type(purpose))
    if not isinstance(length, six.integer_types): raise TypeError(type(length))
    return HKDF(key, length, CTXinfo=purpose)

def derive_phase_key(key, side, phase):
    assert isinstance(side, type("")), type(side)
    assert isinstance(phase, type("")), type(phase)
    side_bytes = side.encode("ascii")
    phase_bytes = phase.encode("ascii")
    purpose = (b"wormhole:phase:"
               + sha256(side_bytes).digest()
               + sha256(phase_bytes).digest())
    return derive_key(key, purpose)

def decrypt_data(key, encrypted):
    assert isinstance(key, type(b"")), type(key)
    assert isinstance(encrypted, type(b"")), type(encrypted)
    assert len(key) == SecretBox.KEY_SIZE, len(key)
    box = SecretBox(key)
    data = box.decrypt(encrypted)
    return data

def encrypt_data(key, plaintext):
    assert isinstance(key, type(b"")), type(key)
    assert isinstance(plaintext, type(b"")), type(plaintext)
    assert len(key) == SecretBox.KEY_SIZE, len(key)
    box = SecretBox(key)
    nonce = utils.random(SecretBox.NONCE_SIZE)
    return box.encrypt(plaintext, nonce)

# the Key we expose to callers (Boss, Ordering) is responsible for sorting
# the two messages (got_code and got_pake), then delivering them to
# _SortedKey in the right order.

@attrs
@implementer(_interfaces.IKey)
class Key(object):
    _appid = attrib(validator=instance_of(type(u"")))
    _versions = attrib(validator=instance_of(dict))
    _side = attrib(validator=instance_of(type(u"")))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)

    def __attrs_post_init__(self):
        self._SK = _SortedKey(self._appid, self._versions, self._side,
                              self._timing)
        self._debug_pake_stashed = False # for tests

    def wire(self, boss, mailbox, receive):
        self._SK.wire(boss, mailbox, receive)

    @m.state(initial=True)
    def S00(self): pass # pragma: no cover
    @m.state()
    def S01(self): pass # pragma: no cover
    @m.state()
    def S10(self): pass # pragma: no cover
    @m.state()
    def S11(self): pass # pragma: no cover

    @m.input()
    def got_code(self, code): pass
    @m.input()
    def got_pake(self, body): pass

    @m.output()
    def stash_pake(self, body):
        self._pake = body
        self._debug_pake_stashed = True
    @m.output()
    def deliver_code(self, code):
        self._SK.got_code(code)
    @m.output()
    def deliver_pake(self, body):
        self._SK.got_pake(body)
    @m.output()
    def deliver_code_and_stashed_pake(self, code):
        self._SK.got_code(code)
        self._SK.got_pake(self._pake)

    S00.upon(got_code, enter=S10, outputs=[deliver_code])
    S10.upon(got_pake, enter=S11, outputs=[deliver_pake])
    S00.upon(got_pake, enter=S01, outputs=[stash_pake])
    S01.upon(got_code, enter=S11, outputs=[deliver_code_and_stashed_pake])

@attrs
class _SortedKey(object):
    _appid = attrib(validator=instance_of(type(u"")))
    _versions = attrib(validator=instance_of(dict))
    _side = attrib(validator=instance_of(type(u"")))
    _timing = attrib(validator=provides(_interfaces.ITiming))
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)

    def wire(self, boss, mailbox, receive):
        self._B = _interfaces.IBoss(boss)
        self._M = _interfaces.IMailbox(mailbox)
        self._R = _interfaces.IReceive(receive)

    @m.state(initial=True)
    def S0_know_nothing(self): pass # pragma: no cover
    @m.state()
    def S1_know_code(self): pass # pragma: no cover
    @m.state()
    def S2_know_key(self): pass # pragma: no cover
    @m.state(terminal=True)
    def S3_scared(self): pass # pragma: no cover

    # from Boss
    @m.input()
    def got_code(self, code): pass

    # from Ordering
    def got_pake(self, body):
        assert isinstance(body, type(b"")), type(body)
        payload = bytes_to_dict(body)
        if "pake_v1" in payload:
            self.got_pake_good(hexstr_to_bytes(payload["pake_v1"]))
        else:
            self.got_pake_bad()
    @m.input()
    def got_pake_good(self, msg2): pass
    @m.input()
    def got_pake_bad(self): pass

    @m.output()
    def build_pake(self, code):
        with self._timing.add("pake1", waiting="crypto"):
            self._sp = SPAKE2_Symmetric(to_bytes(code),
                                        idSymmetric=to_bytes(self._appid))
            msg1 = self._sp.start()
        body = dict_to_bytes({"pake_v1": bytes_to_hexstr(msg1)})
        self._M.add_message("pake", body)

    @m.output()
    def scared(self):
        self._B.scared()
    @m.output()
    def compute_key(self, msg2):
        assert isinstance(msg2, type(b""))
        with self._timing.add("pake2", waiting="crypto"):
            key = self._sp.finish(msg2)
        # TODO: make B.got_key() an eventual send, since it will fire the
        # user/application-layer get_unverified_key() Deferred, and if that
        # calls back into other wormhole APIs, bad things will happen
        self._B.got_key(key)
        phase = "version"
        data_key = derive_phase_key(key, self._side, phase)
        plaintext = dict_to_bytes(self._versions)
        encrypted = encrypt_data(data_key, plaintext)
        self._M.add_message(phase, encrypted)
        # TODO: R.got_key() needs to be eventual-send too, as it can trigger
        # app-level got_verifier() and got_message() Deferreds.
        self._R.got_key(key)

    S0_know_nothing.upon(got_code, enter=S1_know_code, outputs=[build_pake])
    S1_know_code.upon(got_pake_good, enter=S2_know_key, outputs=[compute_key])
    S1_know_code.upon(got_pake_bad, enter=S3_scared, outputs=[scared])



from __future__ import print_function, unicode_literals
import os, time, json
import click
from humanize import naturalsize
from .database import get_db

def abbrev(t):
    if t is None:
        return "-"
    if t > 1.0:
        return "%.3fs" % t
    if t > 1e-3:
        return "%.1fms" % (t*1e3)
    return "%.1fus" % (t*1e6)


def print_event(event):
    event_type, started, result, total_bytes, waiting_time, total_time = event
    followthrough = None
    if waiting_time and total_time:
        followthrough = total_time - waiting_time
    print("%17s: total=%7s wait=%7s ft=%7s size=%s (%s)" %
          ("%s-%s" % (event_type, result),
           abbrev(total_time),
           abbrev(waiting_time),
           abbrev(followthrough),
           naturalsize(total_bytes),
           time.ctime(started),
          ))

def show_usage(args):
    print("closed for renovation")
    return 0

def tail_usage(args):
    if not os.path.exists("relay.sqlite"):
        raise click.UsageError(
            "cannot find relay.sqlite, please run from the server directory"
        )
    db = get_db("relay.sqlite")
    # we don't seem to have unique row IDs, so this is an inaccurate and
    # inefficient hack
    seen = set()
    try:
        while True:
            old = time.time() - 2*60*60
            c = db.execute("SELECT * FROM `usage`"
                           " WHERE `started` > ?"
                           " ORDER BY `started` ASC", (old,))
            for row in c.fetchall():
                event = (row["type"], row["started"], row["result"],
                         row["total_bytes"], row["waiting_time"],
                         row["total_time"])
                if event not in seen:
                    print_event(event)
                    seen.add(event)
            time.sleep(2)
    except KeyboardInterrupt:
        return 0

def count_channels(args):
    if not os.path.exists("relay.sqlite"):
        raise click.UsageError(
            "cannot find relay.sqlite, please run from the server directory"
        )
    db = get_db("relay.sqlite")
    c_list = []
    c_dict = {}
    def add(key, value):
        c_list.append((key, value))
        c_dict[key] = value
    OLD = time.time() - 10*60
    def q(query, values=()):
        return list(db.execute(query, values).fetchone().values())[0]
    add("apps", q("SELECT COUNT(DISTINCT(`app_id`)) FROM `nameplates`"))

    add("total nameplates", q("SELECT COUNT() FROM `nameplates`"))
    add("waiting nameplates", q("SELECT COUNT() FROM `nameplates`"
                                " WHERE `second` is null"))
    add("connected nameplates", q("SELECT COUNT() FROM `nameplates`"
                                  " WHERE `second` is not null"))
    add("stale nameplates", q("SELECT COUNT() FROM `nameplates`"
                              " where `updated` < ?", (OLD,)))

    add("total mailboxes", q("SELECT COUNT() FROM `mailboxes`"))
    add("waiting mailboxes", q("SELECT COUNT() FROM `mailboxes`"
                                " WHERE `second` is null"))
    add("connected mailboxes", q("SELECT COUNT() FROM `mailboxes`"
                                 " WHERE `second` is not null"))

    stale_mailboxes = 0
    for mbox_row in db.execute("SELECT * FROM `mailboxes`").fetchall():
        newest = db.execute("SELECT `server_rx` FROM `messages`"
                            " WHERE `app_id`=? AND `mailbox_id`=?"
                            " ORDER BY `server_rx` DESC LIMIT 1",
                            (mbox_row["app_id"], mbox_row["id"])).fetchone()
        if newest and newest[0] < OLD:
            stale_mailboxes += 1
    add("stale mailboxes", stale_mailboxes)

    add("messages", q("SELECT COUNT() FROM `messages`"))

    if args.json:
        print(json.dumps(c_dict))
    else:
        for (key, value) in c_list:
            print(key, value)
    return 0

def count_events(args):
    if not os.path.exists("relay.sqlite"):
        raise click.UsageError(
            "cannot find relay.sqlite, please run from the server directory"
        )
    db = get_db("relay.sqlite")
    c_list = []
    c_dict = {}
    def add(key, value):
        c_list.append((key, value))
        c_dict[key] = value
    def q(query, values=()):
        return list(db.execute(query, values).fetchone().values())[0]

    add("apps", q("SELECT COUNT(DISTINCT(`app_id`)) FROM `nameplate_usage`"))

    add("total nameplates", q("SELECT COUNT() FROM `nameplate_usage`"))
    add("happy nameplates", q("SELECT COUNT() FROM `nameplate_usage`"
                              " WHERE `result`='happy'"))
    add("lonely nameplates", q("SELECT COUNT() FROM `nameplate_usage`"
                               " WHERE `result`='lonely'"))
    add("pruney nameplates", q("SELECT COUNT() FROM `nameplate_usage`"
                               " WHERE `result`='pruney'"))
    add("crowded nameplates", q("SELECT COUNT() FROM `nameplate_usage`"
                                " WHERE `result`='crowded'"))

    add("total mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"))
    add("happy mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                             " WHERE `result`='happy'"))
    add("scary mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                             " WHERE `result`='scary'"))
    add("lonely mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                              " WHERE `result`='lonely'"))
    add("errory mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                              " WHERE `result`='errory'"))
    add("pruney mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                              " WHERE `result`='pruney'"))
    add("crowded mailboxes", q("SELECT COUNT() FROM `mailbox_usage`"
                               " WHERE `result`='crowded'"))

    add("total transit", q("SELECT COUNT() FROM `transit_usage`"))
    add("happy transit", q("SELECT COUNT() FROM `transit_usage`"
                           " WHERE `result`='happy'"))
    add("lonely transit", q("SELECT COUNT() FROM `transit_usage`"
                            " WHERE `result`='lonely'"))
    add("errory transit", q("SELECT COUNT() FROM `transit_usage`"
                            " WHERE `result`='errory'"))

    add("transit bytes", q("SELECT SUM(`total_bytes`) FROM `transit_usage`"))

    if args.json:
        print(json.dumps(c_dict))
    else:
        for (key, value) in c_list:
            print(key, value)
    return 0



from __future__ import print_function, absolute_import, unicode_literals

def handle_welcome(welcome, relay_url, my_version, stderr):
    if "motd" in welcome:
        motd_lines = welcome["motd"].splitlines()
        motd_formatted = "\n ".join(motd_lines)
        print("Server (at %s) says:\n %s" % (relay_url, motd_formatted),
              file=stderr)

    # Only warn if we're running a release version (e.g. 0.0.6, not
    # 0.0.6+DISTANCE.gHASH). Only warn once.
    if ("current_cli_version" in welcome
        and "+" not in my_version
        and welcome["current_cli_version"] != my_version):
        print("Warning: errors may occur unless both sides are running the same version", file=stderr)
        print("Server claims %s is current, but ours is %s"
              % (welcome["current_cli_version"], my_version),
              file=stderr)



from setuptools import setup

import versioneer

commands = versioneer.get_cmdclass()

setup(name="magic-wormhole",
      version=versioneer.get_version(),
      description="Securely transfer data between computers",
      author="Brian Warner",
      author_email="warner-magic-wormhole@lothar.com",
      license="MIT",
      url="https://github.com/warner/magic-wormhole",
      package_dir={"": "src"},
      packages=["wormhole",
                "wormhole.cli",
                "wormhole.server",
                "wormhole.test",
                ],
      package_data={"wormhole.server": ["db-schemas/*.sql"]},
      entry_points={
          "console_scripts":
          [
              "wormhole = wormhole.cli.cli:wormhole",
              "wormhole-server = wormhole.server.cli:server",
          ]
      },
      install_requires=[
          "spake2==0.8", "pynacl",
          "six",
          "attrs >= 16.3.0", # 16.3.0 adds __attrs_post_init__
          "twisted[tls] >= 17.5.0", # 17.5.0 adds failAfterFailures=
          "autobahn[twisted] >= 0.14.1",
          "automat",
          "hkdf",
          "tqdm >= 4.13.0", # 4.13.0 fixes crash on NetBSD
          "click",
          "humanize",
          "ipaddress",
          "txtorcon >= 0.19.3",
      ],
      extras_require={
          ':sys_platform=="win32"': ["pypiwin32"],
          "dev": ["mock", "tox", "pyflakes",
                  "magic-wormhole-transit-relay==0.1.0"],
      },
      test_suite="wormhole.test",
      cmdclass=commands,
      )



from __future__ import print_function, absolute_import, unicode_literals
from zope.interface import implementer
from automat import MethodicalMachine
from . import _interfaces

@implementer(_interfaces.ITerminator)
class Terminator(object):
    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)

    def __init__(self):
        self._mood = None

    def wire(self, boss, rendezvous_connector, nameplate, mailbox):
        self._B = _interfaces.IBoss(boss)
        self._RC = _interfaces.IRendezvousConnector(rendezvous_connector)
        self._N = _interfaces.INameplate(nameplate)
        self._M = _interfaces.IMailbox(mailbox)

    # 4*2-1 main states:
    # (nm, m, n, 0): nameplate and/or mailbox is active
    # (o, ""): open (not-yet-closing), or trying to close
    # S0 is special: we don't hang out in it

    # TODO: rename o to 0, "" to 1. "S1" is special/terminal
    # so S0nm/S0n/S0m/S0, S1nm/S1n/S1m/(S1)
    
    # We start in Snmo (non-closing). When both nameplate and mailboxes are
    # done, and we're closing, then we stop the RendezvousConnector

    @m.state(initial=True)
    def Snmo(self): pass # pragma: no cover
    @m.state()
    def Smo(self): pass # pragma: no cover
    @m.state()
    def Sno(self): pass # pragma: no cover
    @m.state()
    def S0o(self): pass # pragma: no cover

    @m.state()
    def Snm(self): pass # pragma: no cover
    @m.state()
    def Sm(self): pass # pragma: no cover
    @m.state()
    def Sn(self): pass # pragma: no cover
    #@m.state()
    #def S0(self): pass # unused

    @m.state()
    def S_stopping(self): pass # pragma: no cover
    @m.state()
    def S_stopped(self, terminal=True): pass # pragma: no cover

    # from Boss
    @m.input()
    def close(self, mood): pass

    # from Nameplate
    @m.input()
    def nameplate_done(self): pass

    # from Mailbox
    @m.input()
    def mailbox_done(self): pass

    # from RendezvousConnector
    @m.input()
    def stopped(self): pass


    @m.output()
    def close_nameplate(self, mood):
        self._N.close() # ignores mood
    @m.output()
    def close_mailbox(self, mood):
        self._M.close(mood)

    @m.output()
    def ignore_mood_and_RC_stop(self, mood):
        self._RC.stop()
    @m.output()
    def RC_stop(self):
        self._RC.stop()
    @m.output()
    def B_closed(self):
        self._B.closed()

    Snmo.upon(mailbox_done, enter=Sno, outputs=[])
    Snmo.upon(close, enter=Snm, outputs=[close_nameplate, close_mailbox])
    Snmo.upon(nameplate_done, enter=Smo, outputs=[])

    Sno.upon(close, enter=Sn, outputs=[close_nameplate, close_mailbox])
    Sno.upon(nameplate_done, enter=S0o, outputs=[])

    Smo.upon(close, enter=Sm, outputs=[close_nameplate, close_mailbox])
    Smo.upon(mailbox_done, enter=S0o, outputs=[])

    Snm.upon(mailbox_done, enter=Sn, outputs=[])
    Snm.upon(nameplate_done, enter=Sm, outputs=[])

    Sn.upon(nameplate_done, enter=S_stopping, outputs=[RC_stop])
    S0o.upon(close, enter=S_stopping,
             outputs=[close_nameplate, close_mailbox, ignore_mood_and_RC_stop])
    Sm.upon(mailbox_done, enter=S_stopping, outputs=[RC_stop])

    S_stopping.upon(stopped, enter=S_stopped, outputs=[B_closed])
